<xsl:stylesheet version="3.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
 xmlns:xs="http://www.w3.org/2001/XMLSchema">
 <xsl:output omit-xml-declaration="yes" indent="yes"/>
 
 
 <xsl:template match="node()|@*">
  <xsl:sequence select=
  "
 let
         (:
             The purpose of create is to return an empty tree.

             It returns a sequence of functions,
                - the first function represents the root of a tree,
                - the second function represents the left subtree,
                - and the third function represents the right subtree.
             The value of each function is an empty sequence.
         :)
         $create := (
                                function() { () }      (: root :),
                                function() { () }      (: left :),
                                function() { () }      (: right :)
                            ),


         (:
             empty() returns a boolean value,
             indicating whether $tree is empty.

             $tree is empty in either of these two cases:
             1. $tree is the empty sequence (it doesn't contain any functions).
             2. $tree contains a sequence of three functions, but the first
                  function - representing the root - is empty (i.e., if you invoke the
                  first function it returns the empty sequence).
         :)
         $empty := function($tree as function(*)*)
                           {
                                empty($tree) or empty($tree[1]())
                           },


         (:
             root()returns the value of the root node.

             This function takes one argument, $tree. Since $tree
             is represented by a sequence of functions, returning the
             value of the root node actually means returning the value of
             the function that corresponds to the root node.

             If $tree is empty then the empty sequence is returned. Otherwise
             the *value* of executing the first function in $tree  is returned (recall
             that a tree is represented by a sequence of functions, the first
             function representing the root of the tree).

             Note: $tree[1] is the first function whereas
                        $tree[1]() is the *value* of executing the first function.
         :)
         $root := function($tree as function(*)*)
                        {
                              if ($empty($tree))
                                  then ()
                                  else $tree[1]()
                        },


         (:
             left()returns the left subtree.

             This function takes one argument, $tree. Since $tree is represented
             by a sequence of functions, returning the left subtree
             actually means returning the value of executing the function that corresponds
             to the left subtree.

             If $tree is empty then the empty sequence is returned. Otherwise
             the *value* of executing the second function in $tree is returned (recall
             that a tree is represented by a sequence of functions, the second
             function representing the left subtree).

             Note: $tree[2] is the second function whereas
                        $tree[2]() is the *value* of executing the second function.
         :)
         $left := function($tree as function(*)*)
                      {
                            if ($empty($tree)) then ()
                            else
                                  if ($empty($tree[2])) then ()
                                  else $tree[2]()
                      },


         (:
             right() returns the right subtree.

             This function takes one argument, $tree. Since $tree is represented
             by a sequence of functions, returning the right subtree actually
             means returning the value of executing the function that corresponds to the
             right subtree.

             If $tree is empty then the empty sequence is returned. Otherwise
             the *value* of executing the third function in $tree is returned (recall
             that a tree is represented by a sequence of functions, the third
             function representing the right subtree).
         :)
         $right := function($tree as function(*)*)
                        {
                              if ($empty($tree)) then ()
                             else
                                   if ($empty($tree[3])) then ()
                                   else $tree[3]()
                        },


        (:
            As tree-nodes are inserted into a BST, a comparison is made between
            the value being inserted and the value in the tree's root node. How should
            the comparison be done? That depends on the type of the data contained in a tree-node.
            Comparing two integers is different than comparing two tree fragments.
            So, the insert function must be provided with an appropriate comparator. For
            the case of bank transactions, a date comparison is needed.
            Here is an appropriate comparator function:
        :)
        $transaction-less-than-comparator :=
	        function( $arg1  as element(transaction),
	                          $arg2  as element(transaction)
	                        ) as xs:boolean
	        {
	              xs:date($arg1/@date) lt xs:date($arg2/@date)
	        },


         $numeric-less-than-comparator :=
	         function( $arg1 as xs:decimal,
	                           $arg2 as xs:decimal
	                         ) as xs:boolean
	         {
	               $arg1  lt   $arg2
	         },


         (:
             insert() takes a $tree (BST) and a $item and produces a new BST that is the result
             of inserting $item as a tree-node in $tree.  A BST is a persistent (functional) data-type.
             Insertions or deletions are never done “in place” – instead a new BST is produced.

             The new node is inserted at:
             - if $tree is empty then as root node.
             - if the value of $item is less than the value of  the root node then
               the new tree-node is inserted into the left subtree (note the recursive definition)
             - else the new tree-node is inserted into the right subtree (again note the recursive
               definition)

             Here are the steps taken if $tree is empty:
             - $item specified to be the result of  the  root function. That is, the root function,
               if invoked, it returns $item.
             - A left function is created such that, if invoked, will return an empty subtree.
             - A right function is created such that, if invoked, will return an empty subtree.

             The insert function is recursive. Recursion with anonymous functions requires
             a helper function as discussed before.
         :)
         $insert-helper :=
	         function( $tree as function(*)*,
	                          $item as item(),
	                          $less-than-comparator as function(item(), item()) as xs:boolean,
	                          $insert-helper
	                         )
	         {
	         if ($empty($tree)) then
	               (
	                     function() {$item}     (: root :),
	                     function() {()}            (: left :),
	                     function() {()}            (: right :)
	               )
	          else if ($less-than-comparator($item, $root($tree))) then
	               (
	                     function() {$root($tree)}                           	(: root :),

	                     function() {$insert-helper( $left($tree),
	                                                                     $item,
	                                                                     $less-than-comparator,
	                                                                     $insert-helper)
	                                        }   				(: left :),

	                     function() {$right($tree)}                          	(: right :)
	                )
	          else
	               (
	                     function() {$root($tree)}                          	(: root :),
	                     function() {$left($tree)}                           	(: left :),

	                     function() {$insert-helper( $right($tree),
	                                                                     $item,
	                                                                     $less-than-comparator,
	                                                                     $insert-helper)
	                                          }      				  (: right :)
	                )
	         },


         $insert :=
	         function ( $tree as function(*)*,
	                           $item as item(),
	                           $less-than-comparator as function(item(), item()) as xs:boolean
	                         )
	         {
	               $insert-helper($tree, $item, $less-than-comparator, $insert-helper)
	         },


         (:
             print() produces an XML document that represents the BST $tree.

             The XML document produced by  print()  consists of:
             1. A root element that represents the value of executing the root function.
             2. A left element that represents the value of print() on the left subtree
                (note the recursive definition).
             3. A right element that represents the value of print() on the right subtree
                (note the recursive definition).

            The $print-helper function produces the markup
             as a string and then the $printer function converts
             this string into an XML document by calling the  parse-xml() function.
         :)
         $print-helper :=
	         function ( $tree as function(*)*,
	                           $print-helper
	                          )
	                          as xs:string?
	         {
	         if (not($empty($tree))) then
	                 concat('&lt;tree>',
	                                    '&lt;root>',
	                                          $root($tree),
	                                    '&lt;/root>',
	                                    '&lt;left>',
	                                          $print-helper($left($tree),$print-helper),
	                                    '&lt;/left>',
	                                    '&lt;right>',
	                                          $print-helper($right($tree),$print-helper),
	                                    '&lt;/right>',
	                            '&lt;/tree>'
	                           )
	          else ()
	          },


         $print := function ($tree as function(*)*)
                         {parse-xml($print-helper($tree, $print-helper))/*},


         (:
             populate() produces a new tree from its BST argument $tree by inserting
             the value of the head() of the sequence $items onto $tree and then inserting
             the tail() of the sequence $items onto the result.
         :)
         $populate-helper :=
	         function ( $tree as function(*)*,
	                           $items as item()*,
	                           $less-than-comparator as function(item(), item()) as xs:boolean,
	                           $populate-helper
	                         )
	         {
	             if (empty($items)) then $tree
	             else
	                  $populate-helper( $insert($tree, $items[1], $less-than-comparator),
	                                                    $items[position() gt 1],
	                                                    $less-than-comparator,
	                                                    $populate-helper
	                                                  )
	         },


         $populate :=
	         function( $tree as function(*)*,
	                           $items as item()*,
	                           $less-than-comparator as function(item(), item()) as xs:boolean
	                         )
	         {$populate-helper($tree, $items, $less-than-comparator, $populate-helper)},


        (:
           Finally, the solution of the original problem.
           Task: find all the bank transactions in the range 2012-03-15 to 2012-05-15.
         :)
        $find-range-of-transactions-helper :=
	        function( $tree as function(*)*,
	                          $start-date as xs:date,
	                          $end-date as xs:date,
	                          $find-range-of-transactions-helper
	                )
	                as element(transaction)*
	        {
	            if (empty($tree)) then ()
	            else
	                if (xs:date($root($tree)/@date) eq $start-date) then
	                     (
	                            $root($tree),
	                            $find-range-of-transactions-helper
	                        		        ( $right($tree),
	                                   	          $start-date,
	                                      	          $end-date,
	                                     	          $find-range-of-transactions-helper)
	                     )
	                    else
	                        if (xs:date($root($tree)/@date) eq $end-date) then
	                          (
	                                 $find-range-of-transactions-helper
	                             	            ( $left($tree),
	                                       	              $start-date,
	                                      	              $end-date,
	                                       	              $find-range-of-transactions-helper),
	                                 $root($tree)
	                          )
	                        else
	                            if ((xs:date($root($tree)/@date) gt $start-date)
	                                 and
	                                (xs:date($root($tree)/@date) lt $end-date)) then
	                             (
	                                   $find-range-of-transactions-helper
	                                   		( $left($tree),
	                                      	                 $start-date,
	                                      	                 $end-date,
	                                      	                 $find-range-of-transactions-helper),
	                                   $root($tree),
	                                   $find-range-of-transactions-helper
	                                 		( $right($tree),
	                                   	 	  $start-date,
	                                   	 	  $end-date,
	                                   	 	  $find-range-of-transactions-helper)
	                             )
	                            else
	                                if (xs:date($root($tree)/@date) lt $start-date) then
	                                 (
	                                    $find-range-of-transactions-helper
	                                      	               ( $right($tree),
	                                       	                 $start-date,
	                                       	                 $end-date,
	                                       	                 $find-range-of-transactions-helper)
	                                 )
	                                else
	                                    if (xs:date($root($tree)/@date) gt $end-date) then
	                                     (
	                                        $find-range-of-transactions-helper
	                                          	                    ( $left($tree),
	                                           	                      $start-date,
	                                           	                      $end-date,
	                                           	                      $find-range-of-transactions-helper)
	                                     )
	                                    else ()
	         },


         $find-range-of-transactions :=
	         function($tree as function(*)*,
	                          $start-date as xs:date,
	                          $end-date as xs:date
	                         )
	         {
	             $find-range-of-transactions-helper
	         		       ( $tree,
	               	         $start-date,
	                       	         $end-date,
	                       	         $find-range-of-transactions-helper)
	         }


    (:  At last, we finalize this big, outermost *let* clause with a *return* clause
        that expresses the intent of the users of the BST data-type.
        We want to get all transactions in the period:
        15th March 2012 to 15th May 2012.
    :)

return (
              $find-range-of-transactions
                                     ( $populate((), //transaction, $transaction-less-than-comparator),
                                       xs:date('2012-03-15'),
                                       xs:date('2012-05-15')
                                     )
             )
  "/>
 </xsl:template>
</xsl:stylesheet>